/*
 * Tabs are 4 symbols wide due to stupid Windows function names...
 * ---------------------------------------------------------------
 *
 * Copyright (c) 2010 KSI Labs, LLC. All rights reserved.
 * Written by Sergey Kubushyn <ksi@koi8.net>, <ksi@ksilabs.com>.
 *
 * File: hardware.h
 *
 * This is hardware-related definitions for WDM driver for KSI Labs, LLC
 * CENTIPEDE PCI Board. Both generic hardware functions and Mach3 specific
 * ones are here because Mach3 specific header file is taken almost verbatim
 * from Mach3 SDK to keep it in sync with Mach3 (only slight modifications
 * were made to make it compatible with Windows DDK C compiler.)
 *
 * Please read documentation or go to http://www.ksilabs.com for further info.
 */

/*
 * 						LICENSE and Terms of Use.
 *
 * First of all, this is NOT GPL or Free Software. It is open so you have
 * all the source code for the Engine. There is also VHDL source code for
 * CENTIPEDE board CPLD that is available from KSI Labs, LLC website,
 * http://www.ksilabs.com.
 *
 * As for this particular code, Windows Mach3 Engine, you are free to use it,
 * make any modifications to customize it for your own needs, fix inevitable
 * bugs, port it to other CNC software and do anything else as long as your
 * derivative work that is based on this code or uses parts or pieces thereof
 * is made to be used with any KSI Labs, LLC products.
 *
 * We do NOT grant you a permission to use any parts of this software in any
 * products that are designed for products not manufactured or distributed
 * by KSI Labs, LLC. We might give such a permission on case-by-case basis
 * but it is not legal to use this software or bits and pieces thereof with
 * products not coming from KSI Labs, LLC without our explicit written
 * permission.
 *
 * The original copyright notices must be included in all derivative works.
 * You do NOT have to ask our permission to make any modifications to this
 * code and you are not under obligation to give us any of your work back
 * as long as you are using it for your own needs.
 *
 * However we do NOT give our permission to distribute your derivative work
 * unless all original copyrights are retained, all the conditions of this
 * license are met (use only with KSI Labs, LLC products) and full source
 * code for such derivative work is available for free to everybody without
 * any further conditions including but not limited to signing any kinds of
 * NDA. We explicitely deny permission to distribute any derivative work
 * based on this code with closed source binary-only fragments or include
 * bits or pieces of this code in such products.
 *
 * CNC software like Mach3 and likes is excluded from these requirements as
 * long as it does not include parts or pieces of this software into the
 * close source proprietary application proper. It can use shared memory
 * buffer, IOCTL calls, etc. to communicate with this Engine or the one based
 * on it and keep all the software making those calls or working on that
 * buffer close source and proprietary as long as those calls/structures are
 * documented and that documentation is freely available to everybody without
 * any further conditions including but no limited to signing any kind of NDA.
 */

#ifndef _HARDWARE_H_
#define _HARDWARE_H_

/*****************/
/* Hardware defs */
/*****************/

/* Register offsets */

/* Output port */
#define CP_DOUT_OFF			0x00

/* Input port */
#define CP_DIN_OFF			0x04

/* DAC THR */
#define CP_DAC_OFF			0x08

/* ADC THR */
#define CP_ADC_OFF			0x0c

/* IRQ stuff */
#define CP_IRQSTAT_RAW_OFF	0x10
#define CP_IRQMASK_OFF		0x14
#define CP_IRQSTAT_OFF		0x18

/* Mach 3 stuff */
#define M_IDX_0_OFF			0x1c
#define M_IDX_1_OFF			0x20
#define M_IDX_2_OFF			0x24
#define M_IDX_3_OFF			0x28
#define M_IDX_4_OFF			0x2c
#define M_IDX_5_OFF			0x30
#define M_CFG_OFF			0x34
#define M_CFG2_OFF			0x38
#define M_CTL_OFF			0x3c
#define M_STAT_OFF			0x40
#define M_FIFO_OFF			0x44
#define M_ENC_OFF			0x48
#define M_SPCNT_OFF			0x4c
#define M_STSP_CTL_OFF		0x50
#define M_CFG3_OFF			0x54

/* These are stubs for extension header lines */
#if 0	/* No ext bus in 2.0 Firmware... */
#define CP_EX_0_31_OFF		0x54
#define CP_EX_32_41_OFF		0x58
#define CP_EX_0_31_DIR_OFF	0x5c
#define CP_EX_32_41_DIR_OFF	0x60
#endif

/* Bits */
#define EX_DIR_IN			0
#define EX_DIR_OUT			1
#define CP_IRQ_DAC			(1 << 0)
#define CP_IRQ_ADC			(1 << 1)
#define CP_IRQ_MACH_TMR		(1 << 2)
#define CP_IRQ_MACH_DONE	(1 << 3)
#define CP_IRQ_MACH_FAULT	(1 << 4)
#define CP_IRQ_MACH_FOVF	(1 << 5)
#define CP_IRQ_MACH_PROBE	(1 << 6)
#define CP_IRQ_MACH_HOME0	(1 << 7)
#define CP_IRQ_MACH_HOME1	(1 << 8)
#define CP_IRQ_MACH_HOME2	(1 << 9)
#define CP_IRQ_MACH_HOME3	(1 << 10)
#define CP_IRQ_MACH_HOME4	(1 << 11)
#define CP_IRQ_MACH_HOME5	(1 << 12)
#define CP_SPINDLE_IDX		(1 << 13)
#define CP_SPINDLE_TIMING	(1 << 14)

/* mach_cfg bits */
#define AX_0_EN				(1 << 0)
#define AX_0_DIR_REV		(1 << 1)
#define AX_0_LIMS_SW_NO		(1 << 2)
#define AX_0_HOME_SW_NO		(1 << 3)
#define AX_0_HOME_SW_EN		(1 << 4)
/**/
#define AX_1_EN				(1 << 5)
#define AX_1_DIR_REV		(1 << 6)
#define AX_1_LIMS_SW_NO		(1 << 7)
#define AX_1_HOME_SW_NO		(1 << 8)
#define AX_1_HOME_SW_EN		(1 << 9)
/**/
#define AX_2_EN				(1 << 10)
#define AX_2_DIR_REV		(1 << 11)
#define AX_2_LIMS_SW_NO		(1 << 12)
#define AX_2_HOME_SW_NO		(1 << 13)
#define AX_2_HOME_SW_EN		(1 << 14)
/**/
#define AX_3_EN				(1 << 15)
#define AX_3_DIR_REV		(1 << 16)
#define AX_3_LIMS_SW_NO		(1 << 17)
#define AX_3_HOME_SW_NO		(1 << 18)
#define AX_3_HOME_SW_EN		(1 << 19)
/**/
#define AX_4_EN				(1 << 20)
#define AX_4_DIR_REV		(1 << 21)
#define AX_4_LIMS_SW_NO		(1 << 22)
#define AX_4_HOME_SW_NO		(1 << 23)
#define AX_4_HOME_SW_EN		(1 << 24)
/**/
#define AX_5_EN				(1 << 25)
#define AX_5_DIR_REV		(1 << 26)
#define AX_5_LIMS_SW_NO		(1 << 27)
#define AX_5_HOME_SW_NO		(1 << 28)
#define AX_5_HOME_SW_EN		(1 << 29)
/**/
#define PROBE_SW_NO			(1 << 30)
#define PROBE_SW_EN			(1 << 31)

/* mach_cfg2 bits */
#define AX_0_STEP_REV		(1 << 0)
#define AX_1_STEP_REV		(1 << 1)
#define AX_2_STEP_REV		(1 << 2)
#define AX_3_STEP_REV		(1 << 3)
#define AX_4_STEP_REV		(1 << 4)
#define AX_5_STEP_REV		(1 << 5)
/**/
#define SP_IDX_EN			(1 << 6)
#define SP_IDX_REV			(1 << 7)
/**/
#define SP_DIRSTEP			(1 << 8)
#define SP_STEP_REV			(1 << 9)
#define SP_DIR_REV			(1 << 10)
/**/
#define IDX_DEBOUNCE_MASK    (0x3ff << 11)
#define IDX_DEBOUNCE_BITS(X) (((X) & 0x3ff) << 11)
#define IDX_DEBOUNCE_VAL(X)  (((X) & (SP_DEBOUNCE_MASK)) >> 11)
/**/
#define TIM_DEBOUNCE_MASK    (0x1ff << 21)
#define TIM_DEBOUNCE_BITS(X) (((X) & 0x1ff) << 21)
#define TIM_DEBOUNCE_VAL(X)  (((X) & (SP_DEBOUNCE_MASK)) >> 21)
/**/
#define SP_TIMING_EN		(1 << 30)
#define SP_TIMING_REV		(1 << 31)

/* mach_cfg3 bits */
#define STEPDIR_4PULSE		(1 << 0)
#define CPUMP_ENABLE		(1 << 1)
#define CPUMP_ALWAYS_ON		(1 << 2)
#define CPUMP_5KHZ			(1 << 3)
#define CPUMP_INV			(1 << 4)
#define WD_ENABLE			(1 << 5)

/* mach_stat bits */
#define AX_0_PLUS_SW_HIT	(1 << 0)
#define AX_0_MINUS_SW_HIT	(1 << 1)
#define AX_0_HOME_SW_HIT	(1 << 2)
/**/
#define AX_1_PLUS_SW_HIT	(1 << 3)
#define AX_1_MINUS_SW_HIT	(1 << 4)
#define AX_1_HOME_SW_HIT	(1 << 5)
/**/
#define AX_2_PLUS_SW_HIT	(1 << 6)
#define AX_2_MINUS_SW_HIT	(1 << 7)
#define AX_2_HOME_SW_HIT	(1 << 8)
/**/
#define AX_3_PLUS_SW_HIT	(1 << 9)
#define AX_3_MINUS_SW_HIT	(1 << 10)
#define AX_3_HOME_SW_HIT	(1 << 11)
/**/
#define AX_4_PLUS_SW_HIT	(1 << 12)
#define AX_4_MINUS_SW_HIT	(1 << 13)
#define AX_4_HOME_SW_HIT	(1 << 14)
/**/
#define AX_5_PLUS_SW_HIT	(1 << 15)
#define AX_5_MINUS_SW_HIT	(1 << 16)
#define AX_5_HOME_SW_HIT	(1 << 17)
/**/
#define PROBE_SW_HIT		(1 << 18)
/**/
#define FIFO_ROOM			(0x01f << 19)
#define FIFO_ROOM_BITS(X)	(((X) & 0x01f) << 19)
#define FIFO_ROOM_VAL(X)	(((X) & (FIFO_ROOM)) >> 19)
#define FIFO_EMPTY			(1 << 24)
#define FIFO_FULL			(1 << 25)
/**/
#define RESERVED_0			(1 << 26)
/**/
#define ST_WD_FULL			(1 << 27)
/**/
#define SPINDLE_STOPPED		(1 << 28)
/**/
#define ST_BUSY				(1 << 29)
#define ST_ESTOP			(1 << 30)
#define ST_RUNNING			(1 << 31)

/* Any Home Switch hit */
#define ST_HOME_HIT		(AX_0_HOME_SW_HIT | AX_1_HOME_SW_HIT |\
						AX_2_HOME_SW_HIT | AX_3_HOME_SW_HIT |\
						AX_4_HOME_SW_HIT | AX_5_HOME_SW_HIT)

/* Any Limit Switch Hit */
#define ST_LIMIT_HIT	(AX_0_PLUS_SW_HIT | AX_0_MINUS_SW_HIT |\
						AX_1_PLUS_SW_HIT | AX_1_MINUS_SW_HIT |\
						AX_2_PLUS_SW_HIT | AX_2_MINUS_SW_HIT |\
						AX_3_PLUS_SW_HIT | AX_3_MINUS_SW_HIT |\
						AX_4_PLUS_SW_HIT | AX_4_MINUS_SW_HIT |\
						AX_5_PLUS_SW_HIT | AX_5_MINUS_SW_HIT)

/* mach_ctl bits */
#define SIMULATE			(1 << 0)
#define ESTOP_REQ			(1 << 1)
#define NO_LIMITS			(1 << 2)
#define MACH_RUN			(1 << 3)
#define MACH_RESET			(1 << 4)
/**/
#define MACH_CYCLE_MASK		(0x7ff << 5)
#define MACH_CYCLE_BITS(X)	(((X) & 0x7ff) << 5)
#define MACH_CYCLE_VAL(X)	(((X) & (MACH_CYCLE_MASK)) >> 5)
/**/
#define FIFO_PURGE			(1 << 30)	/* Not real register bit, no readback */
/**/
#define WAIT4PULSE			(1 << 31)	/* Not real register bit, no readback */
/**/

/* mach_sp_ctl bits */
/**/
#define SP_ST_DIV			0x1ffffff
#define SP_ST_DIV_BITS(X)	((X) & 0x1ffffff)
#define SP_ST_DIV_VAL(X)	((X) & (SP_ST_DIV))
/**/
#define SP_ST_DIR			(1 << 25)
#define SP_ST_RUN			(1 << 26)
/**/
#define FAKE_SP_PULSE		(1 << 27)	/* Not real register bit, no readback */

/* FIFO bits */
#define AX_0_DIR			(1 << 0)
#define AX_0_STEP			(1 << 1)
#define AX_1_DIR			(1 << 2)
#define AX_1_STEP			(1 << 3)
#define AX_2_DIR			(1 << 4)
#define AX_2_STEP			(1 << 5)
#define AX_3_DIR			(1 << 6)
#define AX_3_STEP			(1 << 7)
#define AX_4_DIR			(1 << 8)
#define AX_4_STEP			(1 << 9)
#define AX_5_DIR			(1 << 10)
#define AX_5_STEP			(1 << 11)

/* Inputs (din) mappings for Mach3 signals */
#define AX_0_PLUS_SW		(1 << 0)
#define AX_0_MINUS_SW		(1 << 1)
#define AX_0_HOME_SW		(1 << 2)
/**/
#define AX_1_PLUS_SW		(1 << 3)
#define AX_1_MINUS_SW		(1 << 4)
#define AX_1_HOME_SW		(1 << 5)
/**/
#define AX_2_PLUS_SW		(1 << 6)
#define AX_2_MINUS_SW		(1 << 7)
#define AX_2_HOME_SW		(1 << 8)
/**/
#define AX_3_PLUS_SW		(1 << 9)
#define AX_3_MINUS_SW		(1 << 10)
#define AX_3_HOME_SW		(1 << 11)
/**/
#define AX_4_PLUS_SW		(1 << 12)
#define AX_4_MINUS_SW		(1 << 13)
#define AX_4_HOME_SW		(1 << 14)
/**/
#define AX_5_PLUS_SW		(1 << 15)
#define AX_5_MINUS_SW		(1 << 16)
#define AX_5_HOME_SW		(1 << 17)
/**/
#define PROBE_SW			(1 << 18)
#define MACH_ESTOP_SW		(1 << 19)
/**/
#define SPINDLE_TIMING		(1 << 22)
#define SPINDLE_IDX			(1 << 23)
/**/
#define ENC_0_A				(1 << 24)
#define ENC_0_B				(1 << 25)
/**/
#define ENC_1_A				(1 << 26)
#define ENC_1_B				(1 << 27)
/**/
#define ENC_2_A				(1 << 28)
#define ENC_2_B				(1 << 29)
/**/
#define ENC_3_A				(1 << 30)
#define ENC_3_B				(1 << 31)

#define CP_IRQ_ANY_HOME		(\
	CP_IRQ_MACH_HOME0 | CP_IRQ_MACH_HOME1 | CP_IRQ_MACH_HOME2 |\
	CP_IRQ_MACH_HOME3 | CP_IRQ_MACH_HOME4 | CP_IRQ_MACH_HOME5)

#define CP_IRQ_ALL_IRQS	(CP_IRQ_DAC | CP_IRQ_ADC | CP_IRQ_MACH_TMR |\
	CP_IRQ_MACH_DONE | CP_IRQ_MACH_FAULT | CP_IRQ_MACH_FOVF |\
	CP_IRQ_MACH_PROBE | CP_IRQ_ANY_HOME | CP_SPINDLE_IDX | CP_SPINDLE_TIMING)

typedef UINT32	uint32_t;
typedef UINT32	*PUINT32;
typedef INT32	int32_t;
typedef INT32	*PINT32;
typedef unsigned char	uint8_t;
typedef signed char	int8_t;

#define INT_MAX	0x7fffffff

#define FIFO_DEPTH			20
#define FIFO_BLOCKS			4		/* FIFO_DEPTH / 5 */

#define CPD_PORT			11		/* CENTIPEDE Port Number for Mach3 */
#define PPORT1				1		/* Parallel Port 1 */
#define PPORT2				2		/* Parallel Port 2 */
#define PP_INPUT			4		/* Parallel Port "Mostly Input" flag */

/**/
#define DAC_R0				(1 << 12)
#define DAC_R1				(1 << 15)
#define DAC_HS				(1 << 14)
#define DAC_PWDN			(1 << 13)

#define DAC_A_WORD(X)		(DAC_R1 | DAC_HS | (((X) & 0x3ff) << 2))
#define DAC_B_WORD(X)		(DAC_HS | (((X) & 0x3ff) << 2))
/**/

/**/
#define ADC_IN0				(0x00 << 6)
#define ADC_IN1				(0x02 << 6)
#define ADC_IN2				(0x04 << 6)
#define ADC_IN3				(0x06 << 6)
#define ADC_PWRDWN			(0x08 << 6)
#define ADC_FAST			(0x09 << 6)
#define ADC_SLOW			(0x0a << 6)
#define ADC_CAL_MID			(0x0b << 6)
#define ADC_CAL_GND			(0x0c << 6)
#define ADC_CAL_PWR			(0x0d << 6)
/**/

#define	DAC_A_L_DOUT		0x10
#define	DAC_A_H_DOUT		0x11
#define	DAC_B_L_DOUT		0x12
#define	DAC_B_H_DOUT		0x13

#define	ADC_0_L_DIN			0x20
#define	ADC_0_H_DIN			0x21
#define	ADC_1_L_DIN			0x22
#define	ADC_1_H_DIN			0x23
#define	ADC_2_L_DIN			0x24
#define	ADC_2_H_DIN			0x25
#define	ADC_3_L_DIN			0x26
#define	ADC_3_H_DIN			0x27

#define PINS_0_7			0x30
#define PINS_8_15			0x31
#define PINS_16_23			0x32
#define PINS_24_31			0x33

struct _CPD_REGS
{
	uint32_t	dout;
	uint32_t	din;
	uint32_t	dac;
	uint32_t	adc;
	uint32_t	irq_raw;
	uint32_t	irq_mask;
	uint32_t	irq_stat;
	int32_t		mach_idx[6];
	uint32_t	mach_cfg;
	uint32_t	mach_cfg2;
	uint32_t	mach_ctl;
	uint32_t	mach_stat;
	uint32_t	mach_fifo;
	uint32_t	mach_enc;
	uint32_t	mach_sp_cnt;
	uint32_t	mach_sp_ctl;
	uint32_t	mach_cfg3;
#if 0	/* No ext bus in v.2.0 Firmware... */
	uint32_t	ex_data_low;
	uint32_t	ex_data_hi;
	uint32_t	ex_dir_low;
	uint32_t	ex_dir_hi;
#endif
};

typedef volatile struct _CPD_REGS	CPD_REGS;
typedef volatile struct _CPD_REGS	*PCPD_REGS;

#define CPD_MAP_IOSPACE_LENGTH	sizeof(CPD_REGS)

#define FIFO_FREE	(FIFO_ROOM_VAL(FdoData->CPDRegsBase->mach_stat) / 5)

typedef struct _FDO_DATA *PFDO_DATA;

// hardware.c
NTSTATUS HWSetPower(PFDO_DATA FdoData, DEVICE_POWER_STATE DeviceState);
BOOLEAN HWShutdown(PFDO_DATA FdoData);
NTSTATUS AllocateDeviceResources(IN OUT PFDO_DATA FdoData, IN PIRP Irp);
NTSTATUS FreeDeviceResources(IN OUT PFDO_DATA FdoData);
BOOLEAN CPDInterruptHandler(IN PKINTERRUPT Interupt, IN PVOID ServiceContext);

#ifdef _ISR_USE_DPC_
VOID CPDEnableInterrupts(PFDO_DATA FdoData);
VOID CPDDpcForIsr(PKDPC Dpc, PDEVICE_OBJECT DeviceObject, PIRP Irp, PVOID Context);
#endif


// mach3.c
BOOLEAN MachDoIoctlStartTimer(PFDO_DATA FdoData);
BOOLEAN MachSetTimerFreq(PFDO_DATA FdoData);
NTSTATUS MachClaimParallelPorts(PFDO_DATA FdoData);
VOID MachHandleProbeHit(PFDO_DATA FdoData);
VOID MachHandleHomeHit(PFDO_DATA FdoData);
NTSTATUS MachPlannedMotion(PFDO_DATA FdoData);
NTSTATUS MachJog(PFDO_DATA FdoData);
VOID MachGetInputs(PFDO_DATA FdoData);
VOID MachPutOutputs(PFDO_DATA FdoData);
VOID MachUpdateIndexes(PFDO_DATA FdoData);
VOID MachUpdateSpindle(PFDO_DATA FdoData);
//VOID MachUpdateEncoders(PFDO_DATA FdoData);
VOID MachUpdateMPGs(PFDO_DATA FdoData);
VOID MachUpdatePEClock(PFDO_DATA FdoData);
NTSTATUS MachLeaveEStop(PFDO_DATA FdoData);
VOID MachPrepareHome1stPass(PFDO_DATA FdoData);
VOID MachPrepareDeref(PFDO_DATA FdoData);
NTSTATUS MachPrepareClearSwitches(PFDO_DATA FdoData);
VOID MachStopSpindle(PFDO_DATA FdoData);
VOID MachCompleteHome(PFDO_DATA FdoData);
VOID MachAbortHome(PFDO_DATA FdoData);
VOID MachUpdateADC(PFDO_DATA FdoData);
VOID MachUpdateDAC(PFDO_DATA FdoData);
VOID MachUpdateDataIn(PFDO_DATA FdoData);
VOID MachUpdateDataOut(PFDO_DATA FdoData);
VOID MachCheckTHC(PFDO_DATA FdoData);

/* DoTHC and macros for its return value (fields packed into UINT32) */
typedef union _THCCorrection_tag {
	struct {
		uint8_t		shifter;
		uint8_t		reserved1;
		uint8_t		reserved2;
		int8_t		delta;
	} bits;
	uint32_t		pkg;
} THC_STEP, *PTHC_STEP;

UINT32 MachDoTHC(PFDO_DATA FdoData);
/* DoTHC ends */

#define MK_MOVE	\
				__asm xor	edi, edi\
				__asm rcr	dh, 1\
				__asm rcl	edi, 2\
				__asm rcr	dl, 1\
				__asm rcl	edi, 2\
				__asm rcr	ch, 1\
				__asm rcl	edi, 2\
				__asm rcr	cl, 1\
				__asm rcl	edi, 2\
				__asm rcr	ah, 1\
				__asm rcl	edi, 2\
				__asm rcr	al, 1\
				__asm rcl	edi, 2\
				__asm or	edi, dword ptr [CurDirs]\
				__asm mov	dword ptr [ebx], edi\
				__asm mov	edi, dword ptr [ebx+8]

#define PP_NUM_SIGNAL_PINS	17

#define SW_COMB	Port2Input

#define RESTORE_LIMITS	\
			brd->mach_stat = ST_LIMIT_HIT;\
			tmp = brd->mach_ctl;\
			if (!mach->LimitsOff && !(mach->Homing && mach->SW_COMB))\
				tmp &= ~NO_LIMITS;\
			else\
				tmp |= NO_LIMITS;\
			brd->mach_ctl = tmp;\
			dummy = brd->mach_ctl;

#define PURGE_FIFO		\
			tmp = brd->mach_ctl;\
			tmp |= FIFO_PURGE;\
			brd->mach_ctl = tmp;\
			dummy = brd->mach_ctl;

// sigfuncs.c
typedef VOID (*pGetSigFunc)(VOID);
extern const pGetSigFunc InPPMap[PP_NUM_SIGNAL_PINS * 2];
extern const pGetSigFunc nInPPMap[PP_NUM_SIGNAL_PINS * 2];
extern const pGetSigFunc InCPDMap[32];
extern const pGetSigFunc nInCPDMap[32];

typedef VOID (*pPutSigFunc)(VOID);
extern const pPutSigFunc OutPPMap[PP_NUM_SIGNAL_PINS * 2];
extern const pPutSigFunc OutCPDMap[32];

VOID MapOneInput(PFDO_DATA FdoData, pGetSigFunc *GetSigMap, INT Port,
												INT Pin, BOOLEAN Negated);
VOID MapOneOutput(PFDO_DATA FdoData, pPutSigFunc *PutSigMap,
												INT Port, INT Pin);

#endif	/* _HARDWARE_H_ */
